########################################################
## forecasting paradigm
########################################################

dir.paradigm <- function(setup, parset){
  inputset <- c("ychar", "xchar","iniend","maxLag","hors")
  if(any( a <- (!names(setup) %in% inputset) ))
    stop(inputset[a], " is missing in 'setup'.")
  if("setup" %in% search()) detach(setup)
  
  attach(setup)  
  if(is.null(xchar)) xchar <- ychar
  y <- eval(as.name(ychar))
  if(!is.ts(y)) stop(ychar, " must be of class 'ts'.")
  x <- ts(sapply(xchar, function(z) eval(as.name(z)) ))
  tsp(x) <- tsp(y)  
  iter <- NROW(window(y, start = iniend + c(0,1) )) - max(hors)
  if(iter < 1) stop("With this 'setup' you would have ",iter," iterations.\n")
  
  ## define forecasting error matrix
  ERR <- matrix(NA, ncol = max(hors), nrow = iter)
  colnames(ERR) <- paste("hor", 1:ncol(ERR), sep="")
  ymd <- as.Date(window(y, iniend, iniend + c(0, iter-1)))
  ym <- paste(as.numeric(format(ymd,"%Y")),
              as.numeric(format(ymd,"%m")), sep="-")
  rownames(ERR) <- ym
  
  for(i in 1:iter) { # i = 1
    infset <- window(x, end = iniend + c(0,i-1))
    postts <- window(y, end(infset) + c(0,1), end(infset) + c(0,max(hors)))
    K <- NCOL(infset)

    for(z in hors) { # z = 1 
      y.resp <- window(y, start(infset) + c(0, maxLag + z - 1), end(infset))
      x.lagged <- embed2(infset, maxLag, z)[,(K+1):(K*maxLag+K)]
      estimdat <- cbind(y.resp = as.numeric(y.resp), x.lagged)
      was.dummy <- FALSE
      if(!is.null(parset$dummy))
        if(parset$dummy){
          crdm <- create.dummy(y.resp)
          dummy <- crdm$dummy
          nextdummy <- crdm$nextdummy
          estimdat <- cbind(estimdat, dummy)
          was.dummy <- TRUE
        }
      
      ## estimation 
      if(parset$method %in% "VAR"){
        parset$hor = z
        fit <- dir.default.estim(estimdat = infset, parset) # (!)
      } else
      fit <- dir.default.estim(estimdat = estimdat, parset)
      print(i)
      print(coef(fit))
      
      ## forecast
      addednewline <- embed2(rbind(infset, rep(NA,K)), maxLag, z)
      newx <- tail(addednewline, 1)[ ,(K+1):(K*maxLag+K), drop = FALSE]
      if(was.dummy)
        newx$dummy = nextdummy
      fcst <- dir.default.fcst(fit, newx, parset)
#      print(newx)
      ERR[i,z] <- postts[z] - fcst  
    }
  }
  detach(setup)
  finERR <- ERR[, setup$hors, drop = FALSE]
  res <- list(ERR = finERR)
  res$setup <- setup
  res$parset <- parset
  res$MSE <- apply(finERR^2, 2, mean)
  res
}



########################################################
## default parameter-set for the forecasting methods
## estimdat - data.frame
## method - character
########################################################
dir.default.estim <- function(estimdat, parset){
  alowedmethods <- c("glmboost","gamboost","VAR", "bruto", "mars")
  method <- parset$method
  if(!method %in% alowedmethods)
    stop("Unknown method \"",method,"\"\n")

  ##----------------------------
  ## glmboost
  ##----------------------------
  if(method %in% "glmboost"){
    ## default parameter set
    definput <- c("mstop", "dummy")
    plausibility.check(parset, definput)
    
    ## fit
    ctrl <- boost_control(mstop = parset$mstop)
    if(parset$dummy)
      mbfit <- glmboost(y.resp ~. , data = estimdat, control = ctrl,
                        contrasts = list(dummy = "contr.sum")) else
    mbfit <- glmboost(y.resp ~. , data = estimdat, control = ctrl)
#    print(mbfit)

#    print(coef(mbfit))
    aic <- AIC(mbfit)
    print(aic)    
    fit <- mbfit[mstop(aic)]
    cat(mstop(aic),"\n")
  }

  ##----------------------------
  ## gamboost
  ##----------------------------
  if(method %in% "gamboost"){
    ## default parameter set
    definput <- c("df", "knots","mstop", "dummy")
    plausibility.check(parset, definput)
    
    ## fit
    ctrl <- boost_control(mstop = parset$mstop)

    if(parset$dummy)
      model <- bbsformula(estimdat, df = parset$df, knots = parset$knots, contr = "dummy") else
    model <- bbsformula(estimdat, df = parset$df, knots = parset$knots)

##    print(model)
    gbfit <- gamboost(model, data = estimdat, control = ctrl)
##    print(fitted(gamboost(model, data = estimdat, control = ctrl)))
    print(coef(gbfit))
    
#    layout(matrix(1:3,ncol=3))
#    plot(gbfit[mstop(AIC(gbfit, "corrected"))],ask=FALSE)
    aic <- AIC(gbfit)
    fit <- gbfit[mstop(aic)]
#    print(jan)
     print(gbfit[mstop(aic)])
### print(aic)
    cat(mstop(aic),"\n")
  }
  
  
  ##----------------------------
  ## VAR
  ##----------------------------
  if(method %in% "VAR"){
    ## default parameter set
    definput <- c("maxLag","restrict","hor","iter")    
    plausibility.check(parset, definput)
    
    ## fit
    hor <- parset$hor
    if(parset$iter)
      estLag <- VARselect(estimdat, lag.max = parset$maxLag) else
    estLag  <- dirVARselect(estimdat, lag.max = parset$maxLag, hor = hor)
    p.aic <- estLag$selection["AIC(n)"]
    if(parset$iter)
      model.aic <- VAR(estimdat, p = p.aic, type="const") else
    model.aic <- dirVAR(estimdat, p.aic, hor = hor, type = "const")
    if(parset$restrict)
      fit <- restrict(model.aic) else
    fit <- model.aic
  }

  ##----------------------------
  ## bruto
  ##----------------------------
  if(method %in% "bruto"){
    ## default parameter set
    definput <- c("cost")    
    plausibility.check(parset, definput)
    
    ## fit
    x <- estimdat[,-1, drop = FALSE]
    y <- estimdat[,1, drop = FALSE]
    if(parset$cost == "log")
      fit <- bruto(x = x, y = y, cost = log(NROW(x))) else
    fit <- bruto(x = x, y = y, cost = parset$cost)
##  print(fit)
  }

  ##----------------------------
  ## mars
  ##----------------------------
  if(method %in% "mars"){
    ## default parameter set
    definput <- c("penalty")    
    plausibility.check(parset, definput)

    ## fit
    x <- estimdat[,-1, drop = FALSE]
    y <- estimdat[,1, drop = FALSE]
    if(parset$penalty == "log")
      fit <- mars(x = x, y = y, penalty = log(NROW(x))) else
    fit <- mars(x = x, y = y, penalty = parset$penalty)
  }

  fit
  
}

############################################################################
## Adjusted Diebold-Mariano Test, (Harvey, Leyborn and Newbold (1997)) 
############################################################################
diebold.mariano.test <- function(x, hypothesis = c("two.sided", "less", "greater"), k, h = 1)
{
  if (NCOL(x) > 1)
    stop("x is not a vector or univariate time series")
  if (any(is.na(x)))
    stop("NAs in x")
    hypothesis <- match.arg(hypothesis)
  DNAME <- deparse(substitute(x))
  n <- NROW(x)
  cv <- acf(x, lag.max = k, type="covariance", plot=FALSE)$acf[,,1]
  eps <- 1.0e-8
  vr <- max(eps, sum(c(cv[1], 2 * cv[-1])) / n) # eq.(5)
  STATISTIC <- mean(x) / sqrt(vr)
  K <- ((n + 1 - 2 * h + (h/n) * (h-1))/n)^(1/2) # eq.(8)
  STATISTIC <- STATISTIC * K
  names(STATISTIC) <- paste("t with",n-1,"degrees of freedom")
  METHOD <- "Modified Diebold-Mariano Test"
  if (hypothesis == "two.sided")
    PVAL <- 2 * pt(-abs(STATISTIC), df=n-1)
  else if (hypothesis == "greater")
    PVAL <- pt(STATISTIC,df=n-1)
  else if (hypothesis == "less")
    PVAL <- pt(STATISTIC, df=n-1, lower.tail = FALSE)
  PARAMETER <- k
  names(PARAMETER) <- "Truncation lag"
  structure(list(statistic = STATISTIC, 
                 parameter = PARAMETER, 
                 hypothesis = hypothesis,
                 p.value = PVAL, 
                 method = METHOD, 
                 data.name = DNAME),
            class = "htest")
}

############################################################################
## Adjusted Multivariate Diebold-Mariano Test (written by rbz)
## Harvey and Newbold (2000), Tests for Multiple Forecast Encompassing
############################################################################
diebold.mariano.mult <- function(y, X, h = 1, alpha = 2)
{
  if(any(is.na(X))) stop("NAs in x")
  DNAME <- deparse(substitute(X))
  n <- NROW(X)
  K <- NCOL(X)
  u <- if(alpha == 1) resid(lm(y ~ X)) else y
  Q1 <- sapply(1:K, function(j) colSums(X[,j] * X * u^2))
  Q2 <- matrix(0, K, K)
  for(m in seq_len(h-1)){
    d <- n - m
    Q2 <- Q2 + sapply(1:K, function(j)
                      colSums(tail(X[,j], d) * head(X, d) * tail(u, d) * head(u, d)))
  }  
  
  Phi <- Q1 + t(Q2) + Q2                # equations (9)-(10), p 473
  Fa <- (K-1)^(-1) * t(y) %*% X %*% solve(Phi) %*% t(X) %*% y # eq. (10)
  ## const <- n * (n - 1) * (n - K + 1)^(-1) * (n + 1 - 2 * h + h*(h-1)/n )^(-1)
  ## MS <- Fa / const                      # bottom of page 474

  STATISTIC <- Fa
  PARAMETER <- c(K-1, n - K + 1)
  PVAL <- pf(STATISTIC, K-1, n - K + 1)
  PVAL <- 2 * min(PVAL, 1 - PVAL)
  METHOD <- "Modified Diebold-Mariano-Type Test"  
  names(STATISTIC) <- paste("F")
  names(PARAMETER) <- c("num df", "denom df")
  RVAL <- list(statistic = STATISTIC,
               parameter = PARAMETER,
               p.value = PVAL, 
               hypothesis = "two.sided",
               method = METHOD,
               data.name = DNAME)
  attr(RVAL, "class") <- "htest"
  return(RVAL)
}
## Example:
## n <- 30
## dat <- sapply(1:3, function(i) rnorm(n, sd = 1))
## dat <- cbind(rnorm(n, sd = 0.1), dat)
## i <- 1
## y <- dat[,i]
## X <- dat[,i] - dat[,-i]
## diebold.mariano.mult(y, X, h = 2)



########################################################
## default  forecasting
########################################################
dir.default.fcst <- function(fit, newx, parset){
  alowedmethods <- c("glmboost", "gamboost","VAR", "bruto", "mars")
  method <- parset$method
  if(!method %in% alowedmethods)
    stop("Unknown method \"",method,"\"\n")
  
  ## mboost, bruto -----------
  if(method %in% c("glmboost", "gamboost", "bruto", "mars"))
    fcst <- predict(fit, newx)

  ## VAR ---------------------
  ## (!) not optimal for 'iter==TRUE'
  if(method %in% "VAR"){
    h = parset$hor
    if(parset$iter)
      fcst <- predict(fit, n.ahead = h)$fcst[[1]][h,1] else
    fcst <- predict(fit, n.ahead = 1)$fcst[[1]][,1]
  }

  fcst
 
}


########################################################
## plausibility check of the parameters
########################################################
plausibility.check <- function(parset, definput){
  if(!is.list(parset)) stop("Parset must be a 'list'.")
  if(any(a <- !definput %in% names(parset)[-1]))
    stop(paste("\n Parameter '",definput[a],
               "' for method '", parset$method,
               "' is missing in parset.\n"))
  if(any(b <- !names(parset)[-1] %in% definput))
    warning(paste("\n Parameter '", names(parset)[-1][b],
                  "' for method '", parset$method,
                  "' seems to be redundant.\n"))
}


########################################################
## generate setups like 'setup'
########################################################
generate.setup <- function(setup, xchars, keep1xchar = TRUE){
  wsetup <- setup
  xchar1 <- setup$xchar[1]
  res <- list()
  k <- 1
  for(z in xchars) { # z = xchars[1]
    if(keep1xchar)
      wsetup$xchar <- c(xchar1, z) else
    wsetup$xchar <- z
    res[[k]] <- wsetup
    k = k + 1
  }
  res
}


########################################################
## create dummy 
########################################################
create.dummy <- function(y.resp){
  if(is.ts(!y.resp))  stop("'y.resp' must be of class 'ts'.")
  if(frequency(y.resp) == 12){
    dummies <- as.factor(months(seq(as.Date(y.resp)[1], length = length(y.resp) + 1,
                                    by = "month"), abbreviate = TRUE))
    dummy <- dummies[-length(dummies)]
    nextdummy <- tail(dummies,1)
  } 
  if(frequency(y.resp) == 4){
    stop("Quarterly data not implemented yet.\n")
    }
  if(!frequency(y.resp) %in% c(4,12))
    stop("Frequency should be '4' or '12'.")
  res = list(dummy = dummy)
  res$nextdummy = nextdummy
  res
}


########################################################
## pretify results
########################################################
pretify.fcsts <- function(lsfcsts){
  meth <- lsfcsts[[1]]$parset$method
  e <- lsfcsts[[1]]$MSE
  k <- length(e)
  cnt <- 0
  
  numrow <- length(lsfcsts)*k
  summ <- as.data.frame(matrix(NA, ncol = 3, nrow = numrow + numrow/length(e)))
  colnames(summ) <- c("ind", "hor", meth)
  
  for(i in seq(length(lsfcsts))) { # i = 1
    xc <- lsfcsts[[i]]$setup$xchar[-1]
    e <- lsfcsts[[i]]$MSE
    ni <- (k*(i-1)+1):(i*k) + cnt
    summ$ind[ni] <- paste(xc, sep="", collapce="")
    summ$hor[ni] <- names(e)
    summ[ni,3] <- e
    cnt = cnt + 1
  }
  summ
}


## ########################################################
## ## not finished yet
## ########################################################
## boldify <- function(a, what = "min"){
##   if(!is.data.frame(a) | !"xtable" %in% class(a))
##     stop("provide xtable data.frame")
##   dat <- as.data.frame(a)
##   classes <- unlist(lapply(dat, function(z) class(z) ))
##   cs <- sum(classes == "character")
##   dat <- dat[, !classes %in% "character"]
##   if(what == "min")
##     ind <- as.numeric(apply(dat, 1, which.min))
##   i = 1
##   for(j in ind){ # j = 7
##     if(!is.na(j))
##        a[i,(j+2)] <- paste("\\textbf{",dat[i,j],"}")
##     i = i + 1
##   }

##   a

## }
