########################################################
## default parameter-set for the simulation methods
## define formula for the fitting procedure in advance
########################################################

mc.paradigm <- function(setup, parset){
  if("setup" %in% search()) detach(setup)
  attach(setup)
  ## create log file
  logging(logheader(d,n,parset),dsimres)  
  logging(paste("Simulation startet at:",zeit()),dsimres)

  ## define resulting objects
  fin.sc <- matrix(NA, nrow = length(stored.ts), ncol=3)
  colnames(fin.sc) <- rep(parset$method,3)
  rownames(fin.sc) <- stored.ts
  fin.gut <- fin.sc[,1,drop=F]
  fin.GUT <- matrix(NA, nrow = n, ncol = nrow(fin.gut))
  colnames(fin.GUT) <- rownames(fin.sc)
  
  for (z in stored.ts) { #  z = "NLAR1a.210.RData"
    load(paste(repo,z,sep="/")) # load (!)
    sC <- GUT <- rep(NA,n)     
    cat(z, "\n")
    
    for(i in 1:n){ # i=2
      ## data preparation
      work <- simy[[i]]            
      dat <- embed2(work$y,d)       
      y <- dat[,1, drop = FALSE]    
      x <- dat[,-1, drop = FALSE]
      
      ## estimation
      fit <- mc.default.estim(x = x, y = y, parset = parset)
      
      ## evaluation
      prtf <- partial.fit(fit, x = x, y = y)
      sC[i] <- setcheck(work$trueset, attr(prtf,"sset"))
      GUT[i] <- MSPE(prtf, truefct = work$truefct)
      ##cat(sC[i],"\n")
      ##cat(GUT[i],"\n\n")
    }
    
    ## archivation
    setsum <- factor(sC,levels=c("underfitting","correct","overfitting"))
    conclusion <- sapply(levels(setsum), function(k) sum(setsum==k))
    fin.sc[stored.ts %in% z,] <- conclusion
    fin.gut[stored.ts %in% z,] <- median(GUT) # (!) not mean
    fin.GUT[, stored.ts %in% z] <- GUT 
    logging(paste("At:", zeit() ,"computed",z),dsimres)
  }

  logging(paste("Simulation stopped at:",zeit(),"\n-\n-\n-\n"), dsimres)      
  rownames(fin.sc) <- rownames(fin.gut) <- gsub("\\.RData","", stored.ts)
  res <- list(fin.GUT = fin.GUT, fin.sc = fin.sc, fin.gut = fin.gut)
  res
}

########################################################
## default estimation scheme for Monte Carlo
########################################################
mc.default.estim <- function(x, y, parset){
  alowedmethods <- c("gamboost","GAMBoost","mars","bruto","HaY")
  method <- parset$method
  if(!method%in%alowedmethods)
    stop("Unknown method \"",method,"\"\n")

  
  ##----------------------------
  ## gamboost
  ##----------------------------
  if(method %in% "gamboost"){
    ## default input parameters used by this method
    definput <- c("df", "knots","mstop")
    plausibility.check(parset, definput)
    
    ## fit
    estimdat <- cbind(y,x)
    ctrl <- boost_control(mstop = parset$mstop)
    model <- bbsformula(estimdat, df = parset$df, knots = parset$knots)
    gbfit <- gamboost(model, data = estimdat, control = ctrl)
    aic <- AIC(gbfit)
    fit <- gbfit[mstop(aic)]
    cat(mstop(aic),"\n")
  }

  
  ##----------------------------
  ## GAMBoost (still not implemented completely
  ##----------------------------
  if(method %in% "GAMBoost"){
    ## default input parameters used by this method
    definput <- c("lambda")
    plausibility.check(parset, definput)

    ## fit
    fit <- GAMBoost(x = as.matrix(x), y = as.matrix(y),
                    penalty = parset$lambda, bdeg = 3, pdiff = 2,
                    family = gaussian(), sparse.boost = TRUE,
                    calc.se = FALSE, AIC.type = 'corrected')
  }


  ##----------------------------
  ## bruto
  ##----------------------------
  if(method %in% "bruto"){
    ## default input parameters used by this method
    definput <- c("cost")    
    plausibility.check(parset, definput)
    
    ## fit
    if(parset$cost == "log")
      fit <- bruto(x = x, y = y, cost = log(NROW(x))) else
    fit <- bruto(x = x, y = y, cost = parset$cost)
  }

  
  ##----------------------------
  ## mars
  ##----------------------------
  if(method %in% "mars"){
    ## default input parameters used by this method
    definput <- c("penalty")    
    plausibility.check(parset, definput)

    ## fit
    if(parset$penalty == "log")
      fit <- mars(x = x, y = y, penalty = log(NROW(x)), degree = 1) else
    fit <- mars(x = x, y = y, penalty = parset$penalty)
  }


  ##----------------------------
  ## HaY
  ##----------------------------
  if(method %in% "HaY"){
    ## default input parameters used by this method
    definput <- c("Smax", "basis")    
    plausibility.check(parset, definput)
    
    ## fit
    estimdat <- cbind(y, x)
    fit <- stepwise(dat = estimdat, Smax = parset$Smax, basis = parset$basis)
  }
  
  fit
}




########################################################
## extract partial fits
########################################################
partial.fit <- function(fit, x, y) {
  gp <- 200                             # number of grid points
  q <- quantile(y[,1], c(.05, .95))
  xp <- seq(q[1],q[2], length = gp)
  res <- sapply(x, function(i) list(xp = xp, yp = rep(0,gp)),
                simplify = F)

  ##============================
  ## gamboost (OK)
  ##============================
  if ("gamboost" %in% class(fit)) {
    ind <- sort(unique(fit$ensemble[,1]))
    fullSet <- names(x)
    selectedSet <- fullSet[ind]
        
    for(w in ind) {  # w=ind[1]
      xp <- as.data.frame(xp)
      colnames(xp) <- fullSet[w]
      
      ## extract ensembles & calc partial fit
      ## table(fit$ensemble[,1])
      ens <- fit$ensembles[1:mstop(fit)][fit$ensemble[,1] == w]
      H <- sapply(ens, function(g) g[[1]]$predict(newdata = xp))    
      h <- rowSums(H)*fit$control$nu                      
      res[[w]]$yp <- h
    }
    attr(res,"sset") <- selectedSet
  }
  
  ##============================
  ## bruto (OK)
  ##============================
  if("bruto" %in% class(fit)) {
    selectedSet <- names(x)[fit$type!="excluded"]
    ind <- which(names(x)%in%selectedSet)
    Xp <- matrix(sapply(x, mean), gp, ncol(x), byrow=TRUE)

    for(w in ind) { # w=1
      Xp1 <- Xp
      Xp1[,w] <- xp
      yp <- predict(fit, Xp1)
      res[[w]]$yp <- as.numeric(yp)
    }
    attr(res,"sset") <- selectedSet
  }

  
  ##============================
  ## mars (OK)
  ##============================
  if("mars" %in% class(fit)) {
    ## essential for 'mars' 
    showcuts <- function(obj)
      {
        tmp <- as.matrix(obj$cuts[obj$sel, ]) 
        if(NCOL(tmp)>1) dimnames(tmp) <- list(NULL, names(x)) else
        dimnames(tmp) <- list(NULL,"Intercept")
        tmp
      } 
    X <- showcuts(fit)
    selectedSet <- names(x)[apply(zapsmall(X) != 0, 2, any)]
    ind <- which(names(x) %in% selectedSet)
    Xp <- as.data.frame(matrix(sapply(x, mean), gp, ncol(x), byrow=TRUE))

    for(w in ind) { # w=1
      Xp1 <- Xp
      Xp1[,w] <- xp
      yp <- predict(fit, Xp1)
      res[[w]]$yp <- as.numeric(yp)
    }
    attr(res,"sset") <- selectedSet
  }

  ##============================
  ## HaY 
  ##============================
  if("stepwise_HaY" %in% class(fit)) {
    fitmgcv <- fit$fit                  # name it 'fitmgcv', not fit (!)
    fullSet <- colnames(x)
    selectedSet <- fit$set
    ind <- which(names(x) %in% selectedSet)
    
    if(length(ind) > 0){
      Xp <- data.frame(matrix(rep(xp,length(ind)), ncol = length(ind)))
      colnames(Xp) <- selectedSet
      Yp <- predict(fitmgcv, Xp, type = "terms") # Fit
      colnames(Yp) <- colnames(Xp)
      for(w in ind) { # w=1
        res[[w]]$yp <- as.numeric(Yp[ ,fullSet[w]])
      }
    }
    attr(res,"sset") <- selectedSet
  }
  res
}


########################################################
## compute MSPE
## inputs: 'prtf' an object from function 'partial.resid'
##          'truefct' true functions (loaded from RData-file)
########################################################
MSPE <- function(prtf, truefct) {
  d <- length(prtf)
  gut <- rep(NA,d)
  for(l in 1:d){ # l=1
    xp <- prtf[[l]]$xp              
    truefun <- truefct[[l]]             # this is a function
    yp.true <- truefun(xp)              # real y
    yp.fit <- prtf[[l]]$yp              # fitted y

    ## centering
    yp.true <- yp.true - mean(yp.true)
    yp.fit <- yp.fit - mean(yp.fit)     

    ## evaluate
    gut[l] <- mean((yp.fit - yp.true)^2)
  }
  sum(gut)/d
}


########################################################
## INPUT:  'set1' true set.
##         'set2' fitted set.
## Example: set1 = c("lag1","lag2")
##          set21 = c("lag1");          set22 = c("lag1", "lag2")
##          set23 = c("lag1", "lag3");  set24 = c("lag1","lag2","lag3")
##
##          setcheck(set1,set21)
##          setcheck(set1,set22)
##          setcheck(set1,set23)
##          setcheck(set1,set24)
########################################################
setcheck <- function(set1,set2) {
  if( all(set2%in%set1)  &&  all(set1%in%set2) )
    {return("correct"); stop}
  
  if( !all(set2%in%set1)  &&  all(set1%in%set2) )
    {return("overfitting"); stop}
  
  if( all(set2%in%set1)  &&  !all(set1%in%set2) )
    {return("underfitting"); stop}
  
  if( !all(set2%in%set1)  &&  !all(set1%in%set2) )
    {return("underfitting"); stop}
}


########################################################
## plausibility check of the parameters
########################################################
plausibility.check <- function(parset, definput){
  if(!is.list(parset)) stop("Parset must be a 'list'.")
  if(any(a <- !definput %in% names(parset)[-1]))
    stop(paste("\n Parameter '",definput[a],
               "' for method '", parset$method,
               "' is missing in parset.\n"))
  if(any(b <- !names(parset)[-1] %in% definput))
    warning(paste("\n Parameter '", names(parset)[-1][b],
                  "' for method '", parset$method,
                  "' seems to be redundant.\n"))
}


